@@ -62,11 +62,11 @@ static constexpr const char* DEFAULT_SYNTAX_HASKELL =
     "re2c:yych:emit = (.code_model.recursive_functions ? 0 : 1);\n"
     "re2c:yybm = \"yybm\";\n"
     "re2c:yybm:hex = 0;\n"
-    "re2c:yyfill = \"yyfill\";\n"
     "re2c:yystable = \"\"; // deprecated\n"
     "re2c:header = \"\";\n"
     "re2c:eof = -1;\n"
     "re2c:sentinel = -1;\n"
+    "re2c:yyfill = \"yyfill\";\n"
     "re2c:yyfill:enable = 1;\n"
     "re2c:yyfill:parameter = 1;\n"
     "re2c:yyfill:check = 1;\n"
@@ -122,6 +122,7 @@ static constexpr const char* DEFAULT_SYNTAX_HASKELL =
     "re2c:YYCTXMARKER = \"_yyctxmarker\";\n"
     "re2c:YYCURSOR = \"_yycursor\";\n"
     "re2c:YYDEBUG = \"yydebug\";\n"
+    "re2c:YYEND = \"yyend\";\n"
     "re2c:YYFILL = \"yyfill\";\n"
     "re2c:YYGETACCEPT = \"yygetaccept\";\n"
     "re2c:YYGETCOND = \"yygetcond\";\n"
@@ -415,6 +416,11 @@ static constexpr const char* DEFAULT_SYNTAX_HASKELL =
     "        ? YYCURSOR \" >= \" YYLIMIT // YYFILL check can only be used with EOF rule $\n"
     "        : YYLESSTHAN);\n"
     "\n"
+    "code:yyend =\n"
+    "    (.api.generic | .yyfill.enable\n"
+    "        ? YYEND\n"
+    "        : YYCURSOR \" >= \" YYLIMIT);\n"
+    "\n"
     "code:yybm_filter = \"(\" yych \" .&. complement 0xFF) /= 0\";\n"
     "\n"
     "code:yybm_match = \"(\" yybm \" ! (\" offset \" + fromIntegral \" yych \") .&. \" mask \") /= 0\";\n"
