@@ -48,6 +48,7 @@ static constexpr const char* DEFAULT_SYNTAX_RUST =
     "re2c:YYFILL:naked = 0;\n"
     "re2c:YYFN = [\"<undefined-function-name>;<undefined-function-type>\"];\n"
     "re2c:yyfn:sep = \";\";\n"
+    "re2c:yyfn:throw = \"\";\n"
     "re2c:yycond = \"yycond\";\n"
     "re2c:yyctable = \"yyctable\";\n"
     "re2c:yyaccept = \"yyaccept\";\n"
@@ -60,13 +61,13 @@ static constexpr const char* DEFAULT_SYNTAX_RUST =
     "re2c:yych:conversion = 0;\n"
     "re2c:yych:literals = hex;\n"
     "re2c:yych:emit = (.code_model.recursive_functions ? 0 : 1);\n"
-    "re2c:yybm = \"yybm\";\n"
+    "re2c:yybm = \"YYBM\";\n"
     "re2c:yybm:hex = 0;\n"
-    "re2c:yyfill = \"\";\n"
     "re2c:yystable = \"\"; // deprecated\n"
     "re2c:header = \"\";\n"
     "re2c:eof = -1;\n"
     "re2c:sentinel = -1;\n"
+    "re2c:yyfill = \"\";\n"
     "re2c:yyfill:enable = 1;\n"
     "re2c:yyfill:parameter = 1;\n"
     "re2c:yyfill:check = 1;\n"
@@ -88,6 +89,7 @@ static constexpr const char* DEFAULT_SYNTAX_RUST =
     "re2c:debug-output = 0;\n"
     "re2c:computed-gotos = 0;\n"
     "re2c:computed-gotos:threshold = 9;\n"
+    "re2c:computed-gotos:relative = 0;\n"
     "re2c:nested-ifs = 0;\n"
     "re2c:case-insensitive = 0;\n"
     "re2c:case-inverted = 0;\n"
@@ -121,6 +123,7 @@ static constexpr const char* DEFAULT_SYNTAX_RUST =
     "re2c:YYCTXMARKER = (.api.generic ? \"YYCTXMARKER\" : (.api.record ? yyrecord \".\") \"yyctxmarker\");\n"
     "re2c:YYCURSOR = (.api.generic ? \"YYCURSOR\" : (.api.record ? yyrecord \".\") \"yycursor\");\n"
     "re2c:YYDEBUG = \"YYDEBUG\";\n"
+    "re2c:YYEND = \"YYEND\";\n"
     "re2c:YYFILL = \"YYFILL\";\n"
     "re2c:YYGETACCEPT = sigil \"{var}\";\n"
     "re2c:YYGETCOND = \"YYGETCONDITION\"; // historic default\n"
@@ -165,7 +168,7 @@ static constexpr const char* DEFAULT_SYNTAX_RUST =
     "code:const_global = code:const_local;\n"
     "\n"
     "code:array_local =\n"
-    "    topindent \"let \" name \": [\" type \"; \" size \"] = [\" nl indent\n"
+    "    topindent (.const ? \"const \" : \"let \") name \": [\" type \"; \" size \"] = [\" nl indent\n"
     "        [row: topindent [elem{0:-2}: elem \", \"] [elem{-1}: elem \",\"] nl]\n"
     "    dedent topindent \"];\" nl;\n"
     "\n"
@@ -183,6 +186,7 @@ static constexpr const char* DEFAULT_SYNTAX_RUST =
     "code:type_uint = \"usize\";\n"
     "code:type_yybm = \"u8\";\n"
     "code:type_yytarget = <undefined>;\n"
+    "code:type_yyctable = <undefined>;\n"
     "\n"
     "code:cmp_eq = \"==\";\n"
     "code:cmp_ne = \"!=\";\n"
@@ -227,6 +231,8 @@ static constexpr const char* DEFAULT_SYNTAX_RUST =
     "code:continue = topindent \"continue\" (.loop_label ? \" \" label) \";\" nl;\n"
     "\n"
     "code:goto = <undefined>;\n"
+    "code:cgoto = <undefined>;\n"
+    "code:cgoto_data = <undefined>;\n"
     "\n"
     "code:fndecl = <undefined>;\n"
     "\n"
@@ -402,7 +408,14 @@ static constexpr const char* DEFAULT_SYNTAX_RUST =
     "            ? \"(\" YYLIMIT \" - \" YYCURSOR \") < \" need\n"
     "            : YYLIMIT \" <= \" YYCURSOR));\n"
     "\n"
+    "code:yyend =\n"
+    "    (.api.generic | .yyfill.enable\n"
+    "        ? YYEND\n"
+    "        : YYLIMIT \" <= \" YYCURSOR);\n"
+    "\n"
     "code:yybm_filter = yych \" & ~0xFF\";\n"
     "\n"
     "code:yybm_match = \"(\" yybm \"[\" offset \"+\" yych \" as usize] & \" mask \") != 0\";\n"
+    "\n"
+    "code:yytarget_filter = <undefined>;\n"
     ;
