@@ -48,11 +48,10 @@ Adfa::Adfa(Tdfa&& dfa,
         , msg(msg)
         , accepts()
         , lower_char(0)
-        , upper_char(charset.back())
+        , upper_char(charset.back() - 1) // -1 for the fake end-of-input symbol
         , state_count(0)
         , head(nullptr)
         , default_state(nullptr)
-        , eof_state(nullptr)
         , start_state(nullptr)
         , finstates(rules.size(), nullptr)
         , custom_start_label(nullptr)
@@ -61,7 +60,6 @@ Adfa::Adfa(Tdfa&& dfa,
         , mtagnames()
         , mtagvars()
         , def_rule(dfa.def_rule)
-        , eof_rule(dfa.eof_rule)
         , key_size(key)
         , max_fill(0)
         , max_nmatch(0)
@@ -74,7 +72,6 @@ Adfa::Adfa(Tdfa&& dfa,
         , post_rule_action(post_rule_action) {
 
     const size_t nstates = dfa.states.size();
-    const size_t nchars = dfa.nchars;
 
     State** i2s = new State*[nstates];
     for (size_t i = 0; i < nstates; ++i) {
@@ -83,6 +80,8 @@ Adfa::Adfa(Tdfa&& dfa,
 
     State** p = &head;
     for (size_t i = 0; i < nstates; ++i) {
+        size_t nchars = dfa.nchars;
+
         TdfaState* t = dfa.states[i];
         State* s = i2s[i];
 
@@ -91,11 +90,24 @@ Adfa::Adfa(Tdfa&& dfa,
         p = &s->next;
 
         s->rule = t->rule;
-        s->rule_tags = t->tcid[dfa.nchars];
-        s->fall_tags = t->tcid[dfa.nchars + 1];
+        s->rule_tags = t->tcid[nchars];
+        s->fall_tags = t->tcid[nchars + 1];
         s->fill = fill[i];
+        s->fallthru = t->fallthru;
         s->fallback = t->fallback; // see note [fallback states]
 
+        // Get rid of the pseudo-transitions on fake end-of-input symbol,
+        // but store a link to the final state and tags.
+        size_t eof_state = t->arcs[nchars - 1];
+        if (eof_state != Tdfa::NIL) {
+            s->eof_state = i2s[eof_state];
+            s->eof_tags = t->tcid[nchars - 1];
+            // There are no final / fallback tags (by TNFA construction $ is at the end).
+            TdfaState* e = dfa.states[eof_state];
+            CHECK(e->tcid[nchars] == TCID0 && e->tcid[nchars + 1] == TCID0);
+        }
+        --nchars;
+
         bool end = true;
         for (uint32_t c = 0; end && c < nchars; ++c) {
             end &= t->arcs[c] == Tdfa::NIL;
@@ -160,14 +172,20 @@ void Adfa::reorder() {
         State* s = todo.front();
         todo.pop();
         ord.push_back(s);
+        s->linked = true;
         for (uint32_t i = 0; i < s->go.span_count; ++i) {
             State* q = s->go.span[i].to;
-            if(q && done.insert(q).second) {
+            if (q && done.insert(q).second) {
                 todo.push(q);
             }
         }
     }
 
+    for (State* s = head; s; s = s->next) {
+        // Must be a final state reachable only over removed end-of-input transition.
+        if (!s->linked) ord.push_back(s);
+    }
+
     DCHECK(state_count == ord.size());
 
     ord.push_back(nullptr);
@@ -220,6 +238,8 @@ void Adfa::split(State* s) {
     move->go.tags = TCID0; // drop hoisted tags
     move->rule_tags = s->rule_tags;
     move->fall_tags = s->fall_tags;
+    move->eof_state = s->eof_state;
+    move->eof_tags = s->eof_tags;
     add_single_trans(s, move);
 }
 
@@ -315,11 +335,12 @@ void Adfa::find_base_state(const opt_t* opts) {
 // wouldn't need to know tunneling results), but it's much worse for tunneling.
 
 void Adfa::prepare(const opt_t* opts) {
-    // create rule states
+    // create rule states for final states
     for (State* s = head; s; s = s->next) {
-        if (s->rule != Rule::NONE && s->rule != eof_rule) {
+        if (s->rule != Rule::NONE) {
             if (!finstates[s->rule]) {
                 State* n = new State;
+                n->linked = true;
                 n->kind = StateKind::RULE;
                 n->rule = s->rule;
                 finstates[s->rule] = n;
@@ -332,16 +353,18 @@ void Adfa::prepare(const opt_t* opts) {
                 }
             }
         }
-
-        // last state, add end-of-input rule $ if needed (see note [end-of-input rule])
-        // skip zero condition `<>` as it does not have end-of-input rule
-        if (!s->next && opts->fill_eof != NOEOF && cond != ZERO_COND) {
-            eof_state = new State;
-            eof_state->kind = StateKind::RULE;
-            eof_state->rule = eof_rule;
-            finstates[eof_rule] = eof_state;
-            add_state_after(eof_state, s);
-            break;
+    }
+    // fix end-of-input links to point to the newly created rule state
+    for (State* s = head; s; s = s->next) {
+        if (s->eof_state != nullptr) {
+            s->eof_state = finstates[s->eof_state->rule];
+        }
+    }
+    // remove unlinked states
+    for (State* s = head; s; s = s->next) {
+        for (State* t = s->next; t && !t->linked; t = s->next) {
+            s->next = t->next;
+            delete t;
         }
     }
 
@@ -359,15 +382,14 @@ void Adfa::prepare(const opt_t* opts) {
     }
 
     // With end-of-input rule $ there is a default quasi-transition on YYFILL failure, so the
-    // default state is needed if there is at least one final state with at least one outgoing
-    // transition to a non-final state.
+    // default state is needed if there are any "fallthrough" states.
     if (!default_state && opts->fill_eof != NOEOF) {
-        bool have_fallback_states = false;
+        bool have_fallthru = false;
 
         for (State* s = head; s; s = s->next) {
-            have_fallback_states |= s->fallback;
+            have_fallthru |= s->fallthru;
 
-            if (!s->next && have_fallback_states) {
+            if (!s->next && have_fallthru) {
                 default_state = new State;
                 add_state_after(default_state, s);
                 break;
@@ -379,7 +401,6 @@ void Adfa::prepare(const opt_t* opts) {
     if (default_state) {
         for (State* s = head; s; s = s->next) {
             if (s->fallback) {
-                DCHECK(s->rule != eof_rule); // see note [end-of-input rule]
                 s->save = accepts.find_or_add({finstates[s->rule], s->fall_tags});
             }
         }
