@@ -615,6 +615,26 @@ class RenderJmp : public RenderCallback {
     FORBID_COPY(RenderJmp);
 };
 
+class RenderCgoto : public RenderCallback {
+    RenderContext& rctx;
+    const CodeCgoto* code;
+
+  public:
+    RenderCgoto(RenderContext& rctx, const CodeCgoto* code)
+        : rctx(rctx), code(code) {}
+
+    void render_var(StxVarId var) override {
+        switch (var) {
+            case StxVarId::ARRAY: rctx.os << code->array; break;
+            case StxVarId::BASE: rctx.os << code->base; break;
+            case StxVarId::INDEX: rctx.os << code->index; break;
+            default: render_global_var(rctx, var); break;
+        }
+    }
+
+    FORBID_COPY(RenderCgoto);
+};
+
 class RenderFnDef : public RenderCallback {
     RenderContext& rctx;
     const CodeFnDef* code;
@@ -649,6 +669,9 @@ class RenderFnDef : public RenderCallback {
             // explicit void type (like Go). But don't write nullptr, as it truncates the output.
             if (code->type != nullptr) rctx.os << code->type;
             break;
+        case StxVarId::THROW:
+            rctx.os << rctx.opts->fn_throw;
+            break;
         case StxVarId::ARGNAME:
             rctx.os << curr_param->name;
             break;
@@ -1143,6 +1166,14 @@ class RenderArray : public RenderCallback {
         }
     }
 
+    bool eval_cond(StxLOpt opt) override {
+        if (opt == StxLOpt::CONST) {
+            return code->constant;
+        }
+        UNREACHABLE();
+        return false;
+    }
+
     FORBID_COPY(RenderArray);
 };
 
@@ -1413,6 +1444,11 @@ static void render(RenderContext& rctx, const Code* code) {
         rctx.opts->render_code_goto(rctx.os, callback);
         break;
     }
+    case CodeKind::CGOTO: {
+        RenderCgoto callback(rctx, &code->cgoto);
+        rctx.opts->render_code_cgoto(rctx.os, callback);
+        break;
+    }
     case CodeKind::CONTINUE: {
         RenderJmp callback(rctx, code->target);
         rctx.opts->render_code_continue(rctx.os, callback);
