@@ -29,7 +29,6 @@ struct TdfaState {
     size_t rule;
     bool fallthru;
     bool fallback;
-    bool deleted;
 
     explicit TdfaState(size_t nchars)
             : arcs(new size_t[nchars])
@@ -37,8 +36,7 @@ struct TdfaState {
             , tcid(nullptr)
             , rule(Rule::NONE)
             , fallthru(false)
-            , fallback(false)
-            , deleted(false) {
+            , fallback(false) {
         size_t sz = nchars + 2; // +2 for final and fallback epsilon-transitions
         tcmd = new tcmd_t* [sz];
         memset(tcmd, 0, sizeof (tcmd_t*) * sz);
@@ -70,9 +68,8 @@ struct Tdfa {
     tcpool_t tcpool;
     tagver_t maxtagver;
     size_t def_rule;
-    size_t eof_rule;
 
-    Tdfa(DfaAllocator& dfa_alc, size_t charset_bounds, size_t def_rule, size_t eof_rule);
+    Tdfa(DfaAllocator& dfa_alc, size_t charset_bounds, size_t def_rule);
     ~Tdfa();
 
     FORBID_COPY(Tdfa);
@@ -81,7 +78,7 @@ struct Tdfa {
 Ret determinization(
         Tnfa&& nfa, Tdfa& dfa, const opt_t* opts, Msg& msg, const std::string& cond) NODISCARD;
 void minimization(Tdfa& dfa, Minimization type);
-void fillpoints(const Tdfa& dfa, const opt_t* opts, std::vector<size_t>& fill);
+void fillpoints(const Tdfa& dfa, std::vector<size_t>& fill);
 void cutoff_dead_rules(Tdfa& dfa, const opt_t* opts, const std::string& cond, Msg& msg);
 void warn_dead_star_rules(const std::vector<std::unique_ptr<Adfa>>& dfas, Msg& msg);
 void insert_fallback_tags(Tdfa& dfa);
