@@ -37,7 +37,7 @@ Code* emit_skeleton_prolog(Output& output) {
     OutAllocator& alc = output.allocator;
     const char* if_cond;
     CodeList* code, *block, *if_code;
-    CodeArgs* args;
+    CodeParams* params;
 
     code = code_list(alc);
 
@@ -48,12 +48,11 @@ Code* emit_skeleton_prolog(Output& output) {
     append(code, code_textraw(alc, "#include <string.h> /* memcpy */"));
     append(code, code_newline(alc));
 
-    args = code_args(alc);
-    append(args, code_arg(alc, "const char *fname"));
-    append(args, code_arg(alc, "size_t unit"));
-    append(args, code_arg(alc, "size_t padding"));
-    append(args, code_arg(alc, "size_t *pfsize"));
-    append(code, code_fdecl(alc, "static void *read_file", args));
+    params = code_params(alc);
+    append(params, code_param(alc, "fname", "const char*"));
+    append(params, code_param(alc, "unit", "size_t"));
+    append(params, code_param(alc, "padding", "size_t"));
+    append(params, code_param(alc, "pfsize", "size_t*"));
 
     block = code_list(alc);
 
@@ -66,7 +65,7 @@ Code* emit_skeleton_prolog(Output& output) {
     append(block, code_stmt(alc, "FILE *f = fopen(fname, \"rb\")"));
     if_code = code_list(alc);
     append(if_code, code_stmt(alc, "goto error"));
-    append(block, code_if_then_else(alc, "f == NULL", if_code, nullptr, false));
+    append(block, code_if_then_else(alc, "f == NULL", if_code, nullptr));
 
     append(block, code_newline(alc));
 
@@ -81,7 +80,7 @@ Code* emit_skeleton_prolog(Output& output) {
     append(block, code_stmt(alc, "buffer = malloc(unit * (fsize + padding))"));
     if_code = code_list(alc);
     append(if_code, code_stmt(alc, "goto error"));
-    append(block, code_if_then_else(alc, "buffer == NULL", if_code, nullptr, false));
+    append(block, code_if_then_else(alc, "buffer == NULL", if_code, nullptr));
 
     append(block, code_newline(alc));
 
@@ -89,7 +88,7 @@ Code* emit_skeleton_prolog(Output& output) {
     if_cond = "fread(buffer, unit, fsize, f) != fsize";
     if_code = code_list(alc);
     append(if_code, code_stmt(alc, "goto error"));
-    append(block, code_if_then_else(alc, if_cond, if_code, nullptr, false));
+    append(block, code_if_then_else(alc, if_cond, if_code, nullptr));
 
     append(block, code_newline(alc));
 
@@ -104,10 +103,10 @@ Code* emit_skeleton_prolog(Output& output) {
     append(block, code_stmt(alc, "free(buffer)"));
     if_code = code_list(alc);
     append(if_code, code_stmt(alc, "fclose(f)"));
-    append(block, code_if_then_else(alc, "f != NULL", if_code, nullptr, false));
+    append(block, code_if_then_else(alc, "f != NULL", if_code, nullptr));
     append(block, code_stmt(alc, "return NULL"));
 
-    append(code, code_block(alc, block, CodeBlock::Kind::WRAPPED));
+    append(code, code_fndef(alc, "read_file", "static void*", params, block));
 
     return code_block(alc, code, CodeBlock::Kind::RAW);
 }
@@ -148,26 +147,26 @@ static void emit_skeleton_defines(Output& output, CodeList* code, const Adfa& df
     append(code, code_textraw(alc, opts->fill_eof == NOEOF
             ? "#define YYFILL(n) { goto loop_end; }"
             : "#define YYFILL(n) 1 /* fail */"));
+    append(code, code_textraw(alc, "#define YYEND() limit <= cursor"));
     append(code, code_newline(alc));
 }
 
 static void emit_skeleton_function_action(Output& output, CodeList* code, const Adfa& dfa) {
     OutAllocator& alc = output.allocator;
     Scratchbuf& o = output.scratchbuf;
+    const opt_t* opts = output.block().opts;
     const uint64_t norule = rule2key(Rule::NONE, dfa.key_size, dfa.def_rule);
-    CodeArgs* args;
+    CodeParams* params;
     CodeList* if_code, *else_code, *body;
     const char* if_cond, *text;
 
-    args = code_args(alc);
-    append(args, code_arg(alc, "unsigned *pkix"));
-    append(args, code_arg(alc, "const YYKEYTYPE *keys"));
-    append(args, code_arg(alc, "const YYCTYPE *start"));
-    append(args, code_arg(alc, "const YYCTYPE *token"));
-    append(args, code_arg(alc, "const YYCTYPE **cursor"));
-    append(args, code_arg(alc, "YYKEYTYPE rule_act"));
-    text = o.cstr("static int action_").str(dfa.name).flush();
-    append(code, code_fdecl(alc, text, args));
+    params = code_params(alc);
+    append(params, code_param(alc, "pkix", "unsigned*"));
+    append(params, code_param(alc, "keys", "const YYKEYTYPE*"));
+    append(params, code_param(alc, "start", "const YYCTYPE*"));
+    append(params, code_param(alc, "token", "const YYCTYPE*"));
+    append(params, code_param(alc, "cursor", "const YYCTYPE**"));
+    append(params, code_param(alc, "rule_act", "YYKEYTYPE"));
 
     body = code_list(alc);
 
@@ -179,14 +178,14 @@ static void emit_skeleton_function_action(Output& output, CodeList* code, const
     append(body, code_stmt(alc, "*pkix = kix + 3"));
 
     if_cond = o.cstr("rule_exp == ").u64(norule).flush();
+    // Don't use code_fncall here as the format string is too long and needs to be broken on
+    // multiple lines. It's ok because the target is always C/C++.
     if_code = code_list(alc);
-    args = code_args(alc);
-    append(args, code_arg(alc, "stderr"));
-    text = o.cstr("\"warning: lex_").str(dfa.name).cstr(": control flow is undefined\"\n")
-            .cstr("\" for input at position %ld, rerun re2c with '-W'\\n\"").flush();
-    append(args, code_arg(alc, text));
-    append(args, code_arg(alc, "pos"));
-    append(if_code, code_fcall(alc, "fprintf", args, ";"));
+    append(if_code, code_text(alc, o.cstr("fprintf(stderr,").flush()));
+    append(if_code, code_text(alc, o.str(opts->indent_str)
+        .cstr("\"warning: lex_").str(dfa.name).cstr(": control flow is undefined\"").flush()));
+    append(if_code, code_text(alc, o.str(opts->indent_str)
+        .cstr("\" for input at position %ld, rerun re2c with '-W'\\n\");").flush()));
     append(body, code_if_then_else(alc, if_cond, if_code, nullptr));
 
     if_cond = "len_act == len_exp && rule_act == rule_exp";
@@ -194,47 +193,45 @@ static void emit_skeleton_function_action(Output& output, CodeList* code, const
     append(if_code, code_stmt(alc, "const YYKEYTYPE offset = keys[kix]"));
     append(if_code, code_stmt(alc, "*cursor = token + offset"));
     append(if_code, code_stmt(alc, "return 0"));
+    // Don't use code_fncall here as the format string is too long and needs to be broken on
+    // multiple lines. It's ok because the target is always C/C++.
     else_code = code_list(alc);
-    args = code_args(alc);
-    append(args, code_arg(alc, "stderr"));
-    text = o.cstr("\"error: lex_").str(dfa.name).cstr(": at position %ld (key %u):\\n\"\n")
-            .cstr("\"\\texpected: match length %ld, rule %u\\n\"\n")
-            .cstr("\"\\tactual:   match length %ld, rule %u\\n\"").flush();
-    append(args, code_arg(alc, text));
-    append(args, code_arg(alc, "pos"));
-    append(args, code_arg(alc, "kix"));
-    append(args, code_arg(alc, "len_exp"));
-    append(args, code_arg(alc, "rule_exp"));
-    append(args, code_arg(alc, "len_act"));
-    append(args, code_arg(alc, "rule_act"));
-    append(else_code, code_fcall(alc, "fprintf", args, ";"));
+    append(else_code, code_text(alc, "fprintf(stderr,"));
+    append(else_code, code_text(alc, o.str(opts->indent_str)
+        .cstr("\"error: lex_").str(dfa.name).cstr(": at position %ld (key %u):\\n\"").flush()));
+    append(else_code, code_text(alc, o.str(opts->indent_str)
+        .cstr("\"\\texpected: match length %ld, rule %u\\n\"").flush()));
+    append(else_code, code_text(alc, o.str(opts->indent_str)
+        .cstr("\"\\tactual:   match length %ld, rule %u\\n\",").flush()));
+    append(else_code, code_text(alc, o.str(opts->indent_str)
+        .cstr("pos, kix, len_exp, rule_exp, len_act, rule_act);").flush()));
     append(else_code, code_text(alc, "return 1;"));
     append(body, code_if_then_else(alc, if_cond, if_code, else_code));
 
-    append(code, code_block(alc, body, CodeBlock::Kind::WRAPPED));
+    text = o.cstr("action_").str(dfa.name).flush();
+    append(code, code_fndef(alc, text, "static int", params, body));
 }
 
 static void emit_skeleton_stags(Output& output, CodeList* code, const Adfa& dfa) {
     const opt_t* opts = output.block().opts;
-    if (dfa.stagvars.empty() && !opts->tags_posix_syntax) return;
+    if (dfa.stagvars.empty() && !opts->captures) return;
 
     OutAllocator& alc = output.allocator;
     Scratchbuf& o = output.scratchbuf;
     CodeArgs* args;
+    CodeParams* params;
     CodeList* if_code, *body;
     const char* if_cond, *text;
 
     append(code, code_newline(alc));
 
-    args = code_args(alc);
-    append(args, code_arg(alc, "unsigned *pkix"));
-    append(args, code_arg(alc, "YYKEYTYPE *keys"));
-    append(args, code_arg(alc, "const YYCTYPE *tag"));
-    append(args, code_arg(alc, "const YYCTYPE *input"));
-    append(args, code_arg(alc, "const YYCTYPE *token"));
-    append(args, code_arg(alc, "const char *name"));
-    text = o.cstr("static int check_stag_").str(dfa.name).flush();
-    append(code, code_fdecl(alc, text, args));
+    params = code_params(alc);
+    append(params, code_param(alc, "pkix", "unsigned*"));
+    append(params, code_param(alc, "keys", "YYKEYTYPE*"));
+    append(params, code_param(alc, "tag", "const YYCTYPE*"));
+    append(params, code_param(alc, "input", "const YYCTYPE*"));
+    append(params, code_param(alc, "token", "const YYCTYPE*"));
+    append(params, code_param(alc, "name", "const char*"));
 
     body = code_list(alc);
 
@@ -256,22 +253,23 @@ static void emit_skeleton_stags(Output& output, CodeList* code, const Adfa& dfa)
     args = code_args(alc);
     append(args, code_arg(alc, "stderr"));
     text = o.cstr("\"error: lex_").str(dfa.name).cstr(": at position %ld, key %u: \"\n")
-            .cstr("\"wrong value for tag '%s': expected %u, actual %u\\n\"").flush();
+            .cstr("\"wrong value for tag `%s`: expected %u, actual %u\\n\"").flush();
     append(args, code_arg(alc, text));
     append(args, code_arg(alc, "token - input"));
     append(args, code_arg(alc, "kix"));
     append(args, code_arg(alc, "name"));
     append(args, code_arg(alc, "exp"));
     append(args, code_arg(alc, "act"));
-    append(body, code_fcall(alc, "fprintf", args, ";"));
+    append(body, code_fncall(alc, "fprintf", /*retval*/ nullptr, args));
     append(body, code_stmt(alc, "return 1"));
 
-    append(code, code_block(alc, body, CodeBlock::Kind::WRAPPED));
+    text = o.cstr("check_stag_").str(dfa.name).flush();
+    append(code, code_fndef(alc, text, "static int", params, body));
 }
 
 static void emit_skeleton_mtag_defs(Output& output, CodeList* code) {
     OutAllocator& alc = output.allocator;
-    CodeArgs* args;
+    CodeParams* params;
     CodeList* if_code, *block;
     const char* if_cond;
 
@@ -299,40 +297,36 @@ static void emit_skeleton_mtag_defs(Output& output, CodeList* code) {
 
     append(code, code_newline(alc));
 
-    args = code_args(alc);
-    append(args, code_arg(alc, "yymtagpool_t *tp"));
-    append(code, code_fdecl(alc, "static void yymtagpool_clear", args));
+    params = code_params(alc);
+    append(params, code_param(alc, "tp", "yymtagpool_t*"));
     block = code_list(alc);
     append(block, code_stmt(alc, "tp->next = tp->head"));
-    append(code, code_block(alc, block, CodeBlock::Kind::WRAPPED));
+    append(code, code_fndef(alc, "yymtagpool_clear", "static void", params, block));
 
     append(code, code_newline(alc));
 
-    args = code_args(alc);
-    append(args, code_arg(alc, "yymtagpool_t *tp"));
-    append(code, code_fdecl(alc, "static void yymtagpool_init", args));
+    params = code_params(alc);
+    append(params, code_param(alc, "tp", "yymtagpool_t*"));
     block = code_list(alc);
     append(block, code_stmt(alc, "static const unsigned size = 4096"));
     append(block, code_stmt(alc, "tp->head = (yymtag_t*)malloc(size * sizeof(yymtag_t))"));
     append(block, code_stmt(alc, "tp->next = tp->head"));
     append(block, code_stmt(alc, "tp->last = tp->head + size"));
-    append(code, code_block(alc, block, CodeBlock::Kind::WRAPPED));
+    append(code, code_fndef(alc, "yymtagpool_init", "static void", params, block));
 
     append(code, code_newline(alc));
 
-    args = code_args(alc);
-    append(args, code_arg(alc, "yymtagpool_t *tp"));
-    append(code, code_fdecl(alc, "static void yymtagpool_free", args));
+    params = code_params(alc);
+    append(params, code_param(alc, "tp", "yymtagpool_t*"));
     block = code_list(alc);
     append(block, code_stmt(alc, "free(tp->head)"));
     append(block, code_stmt(alc, "tp->head = tp->next = tp->last = NULL"));
-    append(code, code_block(alc, block, CodeBlock::Kind::WRAPPED));
+    append(code, code_fndef(alc, "yymtagpool_free", "static void", params, block));
 
     append(code, code_newline(alc));
 
-    args = code_args(alc);
-    append(args, code_arg(alc, "yymtagpool_t *tp"));
-    append(code, code_fdecl(alc, "static yymtag_t *yymtagpool_next", args));
+    params = code_params(alc);
+    append(params, code_param(alc, "tp", "yymtagpool_t*"));
     block = code_list(alc);
     if_cond = "tp->next == tp->last";
     if_code = code_list(alc);
@@ -346,21 +340,20 @@ static void emit_skeleton_mtag_defs(Output& output, CodeList* code) {
     append(if_code, code_stmt(alc, "tp->last = head + size * 2"));
     append(block, code_if_then_else(alc, if_cond, if_code, nullptr));
     append(block, code_stmt(alc, "return tp->next++"));
-    append(code, code_block(alc, block, CodeBlock::Kind::WRAPPED));
+    append(code, code_fndef(alc, "yymtagpool_next", "static yymtag_t*", params, block));
 
     append(code, code_newline(alc));
 
-    args = code_args(alc);
-    append(args, code_arg(alc, "ptrdiff_t *pt"));
-    append(args, code_arg(alc, "const void *t"));
-    append(args, code_arg(alc, "yymtagpool_t *tp"));
-    append(code, code_fdecl(alc, "static void yymtag", args));
+    params = code_params(alc);
+    append(params, code_param(alc, "pt", "ptrdiff_t*"));
+    append(params, code_param(alc, "t", "const void*"));
+    append(params, code_param(alc, "tp", "yymtagpool_t*"));
     block = code_list(alc);
     append(block, code_stmt(alc, "yymtag_t *n = yymtagpool_next(tp)"));
     append(block, code_stmt(alc, "n->pred = *pt"));
     append(block, code_stmt(alc, "n->elem = t"));
     append(block, code_stmt(alc, "*pt = n - tp->head"));
-    append(code, code_block(alc, block, CodeBlock::Kind::WRAPPED));
+    append(code, code_fndef(alc, "yymtag", "static void", params, block));
 
     append(code, code_newline(alc));
 }
@@ -371,6 +364,7 @@ static void emit_skeleton_mtags(Output& output, CodeList* code, const Adfa& dfa)
     OutAllocator& alc = output.allocator;
     Scratchbuf& o = output.scratchbuf;
     CodeArgs* args;
+    CodeParams* params;
     CodeList* if_code, *block, *body;
     const char* if_cond, *text;
 
@@ -384,16 +378,14 @@ static void emit_skeleton_mtags(Output& output, CodeList* code, const Adfa& dfa)
 
     append(code, code_newline(alc));
 
-    args = code_args(alc);
-    append(args, code_arg(alc, "unsigned *pkix"));
-    append(args, code_arg(alc, "YYKEYTYPE *keys"));
-    append(args, code_arg(alc, "const yymtagpool_t *tp"));
-    append(args, code_arg(alc, "ptrdiff_t mtag"));
-    append(args, code_arg(alc, "const YYCTYPE *input"));
-    append(args, code_arg(alc, "const YYCTYPE *token"));
-    append(args, code_arg(alc, "const char *name"));
-    text = o.cstr("static int check_mtag_").str(dfa.name).flush();
-    append(code, code_fdecl(alc, text, args));
+    params = code_params(alc);
+    append(params, code_param(alc, "pkix", "unsigned*"));
+    append(params, code_param(alc, "keys", "YYKEYTYPE*"));
+    append(params, code_param(alc, "tp", "const yymtagpool_t*"));
+    append(params, code_param(alc, "mtag", "ptrdiff_t"));
+    append(params, code_param(alc, "input", "const YYCTYPE*"));
+    append(params, code_param(alc, "token", "const YYCTYPE*"));
+    append(params, code_param(alc, "name", "const char*"));
 
     body = code_list(alc);
     append(body, code_stmt(alc, "const unsigned kix = *pkix"));
@@ -407,12 +399,12 @@ static void emit_skeleton_mtags(Output& output, CodeList* code, const Adfa& dfa)
     args = code_args(alc);
     append(args, code_arg(alc, "stderr"));
     text = o.cstr("\"error: lex_").str(dfa.name).cstr(": at position %ld, key %u: \"\n")
-            .cstr("\"history for tag '%s' is too short\\n\"").flush();
+            .cstr("\"history for tag `%s` is too short\\n\"").flush();
     append(args, code_arg(alc, text));
     append(args, code_arg(alc, "token - input"));
     append(args, code_arg(alc, "kix + n"));
     append(args, code_arg(alc, "name"));
-    append(if_code, code_fcall(alc, "fprintf", args, ";"));
+    append(if_code, code_fncall(alc, "fprintf", /*retval*/ nullptr, args));
     append(if_code, code_stmt(alc, "return 1"));
     append(block, code_if_then_else(alc, "mtag == -1", if_code, nullptr));
 
@@ -427,14 +419,14 @@ static void emit_skeleton_mtags(Output& output, CodeList* code, const Adfa& dfa)
     args = code_args(alc);
     append(args, code_arg(alc, "stderr"));
     text = o.cstr("\"error: lex_").str(dfa.name).cstr(": at position %ld, key %u: \"\n")
-            .cstr("\"wrong value for tag '%s': expected %u, actual %u\\n\"").flush();
+            .cstr("\"wrong value for tag `%s`: expected %u, actual %u\\n\"").flush();
     append(args, code_arg(alc, text));
     append(args, code_arg(alc, "token - input"));
     append(args, code_arg(alc, "kix + n"));
     append(args, code_arg(alc, "name"));
     append(args, code_arg(alc, "exp"));
     append(args, code_arg(alc, "act"));
-    append(if_code, code_fcall(alc, "fprintf", args, ";"));
+    append(if_code, code_fncall(alc, "fprintf", /*retval*/ nullptr, args));
     append(if_code, code_stmt(alc, "return 1"));
     append(block, code_if_then_else(alc, if_cond, if_code, nullptr));
 
@@ -445,34 +437,34 @@ static void emit_skeleton_mtags(Output& output, CodeList* code, const Adfa& dfa)
     args = code_args(alc);
     append(args, code_arg(alc, "stderr"));
     text = o.cstr("\"error: lex_").str(dfa.name).cstr(": at position %ld, key %u: \"\n")
-            .cstr("\"history for tag '%s' is too long\\n\"").flush();
+            .cstr("\"history for tag `%s` is too long\\n\"").flush();
     append(args, code_arg(alc, text));
     append(args, code_arg(alc, "token - input"));
     append(args, code_arg(alc, "kix + n"));
     append(args, code_arg(alc, "name"));
-    append(if_code, code_fcall(alc, "fprintf", args, ";"));
+    append(if_code, code_fncall(alc, "fprintf", /*retval*/ nullptr, args));
     append(if_code, code_stmt(alc, "return 1"));
     append(body, code_if_then_else(alc, "mtag != -1", if_code, nullptr));
     append(body, code_text(alc, "return 0;"));
 
-    append(code, code_block(alc, body, CodeBlock::Kind::WRAPPED));
+    text = o.cstr("check_mtag_").str(dfa.name).flush();
+    append(code, code_fndef(alc, text, "static int", params, body));
 }
 
 static void emit_skeleton_function_check_key_count(Output& output, CodeList* code, Adfa& dfa) {
     OutAllocator& alc = output.allocator;
     Scratchbuf& o = output.scratchbuf;
     CodeArgs* args;
+    CodeParams* params;
     CodeList* block, *if_code;
     const char* if_cond, *text;
 
     append(code, code_newline(alc));
 
-    args = code_args(alc);
-    append(args, code_arg(alc, "unsigned have"));
-    append(args, code_arg(alc, "unsigned used"));
-    append(args, code_arg(alc, "unsigned need"));
-    text = o.cstr("static int check_key_count_").str(dfa.name).flush();
-    append(code, code_fdecl(alc, text, args));
+    params = code_params(alc);
+    append(params, code_param(alc, "have", "unsigned"));
+    append(params, code_param(alc, "used", "unsigned"));
+    append(params, code_param(alc, "need", "unsigned"));
     block = code_list(alc);
     if_cond = "used + need <= have";
     if_code = code_list(alc);
@@ -482,9 +474,11 @@ static void emit_skeleton_function_check_key_count(Output& output, CodeList* cod
     append(args, code_arg(alc, "stderr"));
     text = o.cstr("\"error: lex_").str(dfa.name).cstr(": not enough keys\\n\"").flush();
     append(args, code_arg(alc, text));
-    append(block, code_fcall(alc, "fprintf", args, ";"));
+    append(block, code_fncall(alc, "fprintf", /*retval*/ nullptr, args));
     append(block, code_stmt(alc, "return 1"));
-    append(code, code_block(alc, block, CodeBlock::Kind::WRAPPED));
+
+    text = o.cstr("check_key_count_").str(dfa.name).flush();
+    append(code, code_fndef(alc, text, "static int", params, block));
 }
 
 static void emit_skeleton_function_lex(Output& output, CodeList* code, Adfa& dfa) {
@@ -505,8 +499,6 @@ static void emit_skeleton_function_lex(Output& output, CodeList* code, Adfa& dfa
 
     append(code, code_newline(alc));
 
-    text = o.cstr("int lex_").str(dfa.name).cstr("()").flush();
-    append(code, code_text(alc, text));
     block = code_list(alc);
     text = o.cstr("const size_t padding = ").u64(dfa.max_fill).cstr("; /* YYMAXFILL */").flush();
     append(block, code_text(alc, text));
@@ -519,7 +511,7 @@ static void emit_skeleton_function_lex(Output& output, CodeList* code, Adfa& dfa
     append(block, code_stmt(alc, "const YYCTYPE *limit = NULL"));
     append(block, code_stmt(alc, "const YYCTYPE *token = NULL"));
     append(block, code_stmt(alc, "const YYCTYPE *eof = NULL"));
-    if (opts->tags_posix_syntax) {
+    if (opts->captures) {
         append(block, code_stmt(alc, "size_t yynmatch"));
         text = o.cstr("const YYCTYPE *yypmatch[").u64(dfa.max_nmatch).cstr(" * 2]").flush();
         append(block, code_stmt(alc, text));
@@ -539,7 +531,7 @@ static void emit_skeleton_function_lex(Output& output, CodeList* code, Adfa& dfa
     append(args, code_arg(alc, "sizeof (YYCTYPE)"));
     append(args, code_arg(alc, "padding"));
     append(args, code_arg(alc, "&input_len"));
-    append(block, code_fcall(alc, "input = (YYCTYPE *) read_file", args, ";"));
+    append(block, code_fncall(alc, "input = (YYCTYPE *) read_file", /*retval*/ nullptr, args));
 
     if_code = code_list(alc);
     append(if_code, code_stmt(alc, "status = 1"));
@@ -564,7 +556,7 @@ static void emit_skeleton_function_lex(Output& output, CodeList* code, Adfa& dfa
     append(args, code_arg(alc, "sizeof (YYKEYTYPE)"));
     append(args, code_arg(alc, "0"));
     append(args, code_arg(alc, "&keys_count"));
-    append(block, code_fcall(alc, "keys = (YYKEYTYPE *) read_file", args, ";"));
+    append(block, code_fncall(alc, "keys = (YYKEYTYPE *) read_file", /*retval*/ nullptr, args));
     if_code = code_list(alc);
     append(if_code, code_stmt(alc, "status = 1"));
     append(if_code, code_stmt(alc, "goto end"));
@@ -611,53 +603,41 @@ static void emit_skeleton_function_lex(Output& output, CodeList* code, Adfa& dfa
     if (dfa.need_accept) {
         append(block2, code_stmt(alc, "unsigned int yyaccept = 0"));
     }
-    if (opts->loop_switch) {
+    if (opts->code_model == CodeModel::LOOP_SWITCH) {
         append(block2, code_stmt(alc, o.cstr("unsigned int yystate = 0").flush()));
     }
     // autogenerated stag variables
     if (!dfa.stagnames.empty()) {
-        text = o.cstr("\n").str(indent(2, opts->indent_str)).cstr("const YYCTYPE *@@ = NULL;")
-                .flush();
-        Code* stags = code_fmt(alc, CodeKind::STAGS, nullptr, text, nullptr);
+        o.cstr("\n").str(indent(2, opts->indent_str)).cstr("const YYCTYPE *@@ = NULL;");
+        Code* stags = code_fmt(alc, CodeKind::STAGS, nullptr, o.flush(), nullptr);
         gen_tags(o, opts, stags, dfa.stagnames);
         append(block2, stags);
         append(block2, code_textraw(alc, ""));
     }
     // user-defined stag variables
     if (!dfa.stagvars.empty()) {
-        var1 = dfa.stagvars.begin();
-        var2 = dfa.stagvars.end();
-        if (var1 != var2) {
-            o.cstr("const YYCTYPE *").str(*var1);
-            for (++var1; var1 != var2; ++var1) {
-                o.cstr(", *").str(*var1);
-            }
-            text = o.flush();
-            append(block2, code_stmt(alc, text));
-        }
+        o.cstr("\n").str(indent(2, opts->indent_str)).cstr("const YYCTYPE *@@ = NULL;");
+        Code* stagvars = code_fmt(alc, CodeKind::SVARS, nullptr, o.flush(), nullptr);
+        gen_tags(o, opts, stagvars, dfa.stagvars);
+        append(block2, stagvars);
+        append(block2, code_textraw(alc, ""));
     }
     // autogenerated mtag variables
     if (!dfa.mtagnames.empty()) {
         append(block2, code_text(alc, "yymtagpool_clear(&yytp);"));
-
-        text = o.cstr("\n").str(indent(2, opts->indent_str)).cstr("ptrdiff_t @@ = -1;").flush();
-        Code* mtags = code_fmt(alc, CodeKind::MTAGS, nullptr, text, nullptr);
+        o.cstr("\n").str(indent(2, opts->indent_str)).cstr("ptrdiff_t @@ = -1;");
+        Code* mtags = code_fmt(alc, CodeKind::MTAGS, nullptr, o.flush(), nullptr);
         gen_tags(o, opts, mtags, dfa.mtagnames);
         append(block2, mtags);
         append(block2, code_textraw(alc, ""));
     }
     // user-defined mtag variables
     if (!dfa.mtagvars.empty()) {
-        var1 = dfa.mtagvars.begin();
-        var2 = dfa.mtagvars.end();
-        if (var1 != var2) {
-            o.cstr("ptrdiff_t ").str(*var1);
-            for (++var1; var1 != var2; ++var1) {
-                o.cstr(", ").str(*var1);
-            }
-            text = o.flush();
-            append(block2, code_stmt(alc, text));
-        }
+        o.cstr("\n").str(indent(2, opts->indent_str)).cstr("ptrdiff_t @@ = -1;");
+        Code* mtagvars = code_fmt(alc, CodeKind::MVARS, nullptr, o.flush(), nullptr);
+        gen_tags(o, opts, mtagvars, dfa.mtagvars);
+        append(block2, mtagvars);
+        append(block2, code_textraw(alc, ""));
     }
 
     if (opts->bitmaps) {
@@ -665,15 +645,22 @@ static void emit_skeleton_function_lex(Output& output, CodeList* code, Adfa& dfa
     }
     append(block2, code_textraw(alc, ""));
 
-    if (!opts->loop_switch) {
+    switch (opts->code_model) {
+    case CodeModel::GOTO_LABEL:
         gen_dfa_as_blocks_with_labels(output, dfa, block2);
-    } else {
-        // With loop/switch there are no labels, and each block has its own state switch. Restart
-        // state counter from zero so that cases start from zero.
+        break;
+    case CodeModel::LOOP_SWITCH: {
+        // In loop/switch mode there are no labels, and each block has its own state switch.
+        // Restart state counter from zero so that cases start from zero.
         output.label_counter = 0;
         CodeCases* cases = code_cases(alc);
         gen_dfa_as_switch_cases(output, dfa, cases);
         wrap_dfas_in_loop_switch(output, block2, cases);
+        break;
+    }
+    case CodeModel::REC_FUNC:
+        UNREACHABLE();
+        break;
     }
     append(block2, code_textraw(alc, ""));
 
@@ -688,16 +675,16 @@ static void emit_skeleton_function_lex(Output& output, CodeList* code, Adfa& dfa
     text = o.cstr("fprintf(stderr, \"error: lex_").str(dfa.name)
             .cstr(": unused input strings left at position %ld\\n\", pos)").flush();
     append(if1_code, code_stmt(alc, text));
-    append(if_code, code_if_then_else(alc, "cursor != eof", if1_code, nullptr, false));
+    append(if_code, code_if_then_else(alc, "cursor != eof", if1_code, nullptr));
 
     if1_code = code_list(alc);
     append(if1_code, code_stmt(alc, "status = 1"));
     text = o.cstr("fprintf(stderr, \"error: lex_").str(dfa.name)
             .cstr(": unused keys left after %u keys\\n\", i)").flush();
     append(if1_code, code_stmt(alc, text));
-    append(if_code, code_if_then_else(alc, "i != keys_count", if1_code, nullptr, false));
+    append(if_code, code_if_then_else(alc, "i != keys_count", if1_code, nullptr));
 
-    append(block, code_if_then_else(alc, "status == 0", if_code, nullptr, false));
+    append(block, code_if_then_else(alc, "status == 0", if_code, nullptr));
 
     append(block, code_textraw(alc, ""));
     append(block, code_slabel(alc, "end"));
@@ -709,7 +696,8 @@ static void emit_skeleton_function_lex(Output& output, CodeList* code, Adfa& dfa
     append(block, code_textraw(alc, ""));
     append(block, code_stmt(alc, "return status"));
 
-    append(code, code_block(alc, block, CodeBlock::Kind::WRAPPED));
+    text = o.cstr("lex_").str(dfa.name).flush();
+    append(code, code_fndef(alc, text, "int", code_params(alc), block));
 }
 
 static void emit_skeleton_undefs(Output& output, CodeList* code, Adfa& dfa) {
@@ -757,14 +745,13 @@ Code* emit_skeleton_epilog(Output& output) {
         const char* if_cond = o.cstr("lex_").str(s).cstr("() != 0").flush();
         CodeList* if_code = code_list(alc);
         append(if_code, code_stmt(alc, "return 1"));
-        append(stmts, code_if_then_else(alc, if_cond, if_code, nullptr, false));
+        append(stmts, code_if_then_else(alc, if_cond, if_code, nullptr));
     }
     append(stmts, code_stmt(alc, "return 0"));
 
     CodeList* main = code_list(alc);
     append(main, code_text(alc, ""));
-    append(main, code_text(alc, "int main()"));
-    append(main, code_block(alc, stmts, CodeBlock::Kind::WRAPPED));
+    append(main, code_fndef(alc, "main", "int", code_params(alc), stmts));
 
     return code_block(alc, main, CodeBlock::Kind::RAW);
 }
@@ -775,8 +762,7 @@ void emit_skeleton_action(Output& output, CodeList* code, const Adfa& dfa, size_
     const std::string& name = dfa.name;
     const Rule& r = dfa.rules[rid];
     const uint64_t rkey = rule2key(rid, dfa.key_size, dfa.def_rule);
-    const char* text;
-    std::vector<std::string> fintags;
+    std::vector<const char*> fintags;
 
     size_t ntag = 3;
     for (size_t t = r.ltag; t < r.htag; ++t) {
@@ -784,48 +770,31 @@ void emit_skeleton_action(Output& output, CodeList* code, const Adfa& dfa, size_
         if (t != r.ttag && !fictive(tag)) ++ntag;
     }
 
-    CodeArgs* args = code_args(alc);
-    append(args, code_arg(alc, "keys_count"));
-    append(args, code_arg(alc, "i"));
-    text = o.u64(ntag).flush();
-    append(args, code_arg(alc, text));
-    text = o.cstr("status = check_key_count_").str(dfa.name).flush();
-    append(code, code_fcall(alc, text, args, ""));
+    // Don't use code_fncall here as it can't handle complex expressions with logical OR of
+    // multiple function calls, and it's always the same target (C/C++).
+    o.cstr("status = check_key_count_").str(dfa.name).cstr("(keys_count, i, ").u64(ntag).cstr(")");
+    append(code, code_text(alc, o.flush()));
 
     CodeList* hangafter = code_list(alc);
 
-    args = code_args(alc);
-    append(args, code_arg(alc, "&i"));
-    append(args, code_arg(alc, "keys"));
-    append(args, code_arg(alc, "input"));
-    append(args, code_arg(alc, "token"));
-    append(args, code_arg(alc, "&cursor"));
-    text = o.u64(rkey).flush();
-    append(args, code_arg(alc, text));
-    text = o.cstr(" || action_").str(name).flush();
-    append(hangafter, code_fcall(alc, text, args, ntag > 3 ? "" : ";"));
+    o.cstr(" || action_").str(name).cstr("(&i, keys, input, token, &cursor, ").u64(rkey).cstr(")");
+    if (ntag <= 3) o.cstr(";");
+    append(hangafter, code_text(alc, o.flush()));
 
     for (size_t t = r.ltag; t < r.htag; ++t) {
         const Tag& tag = dfa.tags[t];
         if (t == r.ttag || fictive(tag)) continue;
 
         const bool mtag = history(tag);
-        expand_fintags(tag, fintags);
+        expand_fintags(output, tag, fintags);
         const std::string tname = fintags[0];
         --ntag;
 
-        args = code_args(alc);
-        append(args, code_arg(alc, "&i"));
-        append(args, code_arg(alc, "keys"));
-        if (mtag) append(args, code_arg(alc, "&yytp"));
-        text = o.str(tname).flush();
-        append(args, code_arg(alc, text));
-        append(args, code_arg(alc, "input"));
-        append(args, code_arg(alc, "token"));
-        text = o.cstr("\"").str(tname).cstr("\"").flush();
-        append(args, code_arg(alc, text));
-        text = o.cstr(" || check_").cstr(mtag ? "mtag_" : "stag_").str(name).flush();
-        append(hangafter, code_fcall(alc, text, args, ntag > 3 ? "" : ";"));
+        o.cstr(" || check_").cstr(mtag ? "mtag_" : "stag_").str(name).cstr("(&i, keys, ");
+        if (mtag) o.cstr("&yytp, ");
+        o.str(tname).cstr(", input, token, ").cstr("\"").str(tname).cstr("\")");
+        if (ntag <= 3) o.cstr(";");
+        append(hangafter, code_text(alc, o.flush()));
     }
 
     append(code, code_block(alc, hangafter, CodeBlock::Kind::INDENTED));
